/*
 * Copyright 2012 Francesco Cappelli
 *
 * This file is part of Pagliaio.
 *
 * Pagliaio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pagliaio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Pagliaio.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <SearchAction.h>
#include <Configuration.h>
#include <PagliaioStandardAnalyzer.h>
#include <PagliaioQueryParser.h>

#include <lucene++/FileUtils.h>
#include <lucene++/QueryTermExtractor.h>
#include <lucene++/WeightedTerm.h>

#include <boost/thread.hpp>
#include <boost/lexical_cast.hpp>

using namespace std;

extern Lucene::IndexReaderPtr indexReader;
extern Lucene::SearcherPtr indexSearcher;
extern boost::shared_mutex indexMutex;
extern pagliaio::Configuration configuration;

namespace pagliaio {

SearchAction::SearchAction() {}

SearchAction::~SearchAction() {}

string SearchAction::generateResponse(void *cls, struct MHD_Connection *connection,
								 	 const char *url,
								 	 const char *method, const char *version,
								 	 const char *upload_data,
								 	 size_t *upload_data_size, void **con_cls)
{
	if (strncmp("GET", method, 3) == 0) {
		map<string, string>* args = new map<string, string>();
		MHD_get_connection_values(connection, MHD_GET_ARGUMENT_KIND, iterate_get, (void*)args);

		if (args->count("query") == 0) {
			addErrorToResponse("Numero o tipo degli argomenti errato.");
		}
		else {
			//make a copy of global indexes
			boost::shared_lock<boost::shared_mutex> lock(indexMutex);
			Lucene::IndexReaderPtr indexReaderLocal = indexReader;
			indexReaderLocal->incRef();
			Lucene::SearcherPtr indexSearcherLocal = indexSearcher;
			lock.unlock();

			//extract optional args
			int maxResults = 100;
			if (args->count("maxResults") != 0)
				maxResults = Lucene::StringUtils::toInt(Lucene::StringUtils::toUnicode((*args)["maxResults"].c_str()));

			Lucene::String defaultField = Lucene::StringUtils::toUnicode(configuration.getDefaultField().c_str());
			//Lucene::AnalyzerPtr analyzer = Lucene::newLucene<Lucene::StandardAnalyzer>(Lucene::LuceneVersion::LUCENE_CURRENT);
			Lucene::AnalyzerPtr analyzer = Lucene::newLucene<PagliaioStandardAnalyzer>(Lucene::LuceneVersion::LUCENE_CURRENT);
			Lucene::QueryParserPtr parser = Lucene::newLucene<PagliaioQueryParser>(Lucene::LuceneVersion::LUCENE_CURRENT, defaultField, analyzer);

			//trim string from initial whitespaces
			string query = string((*args)["query"]);
			query = trim(query);

			Lucene::String queryString = Lucene::StringUtils::toUnicode(query.c_str());
			syslog(LOG_DEBUG, "Parsing query: %ls", queryString.c_str());

			try {
				Lucene::QueryPtr query = parser->parse(queryString);

				Lucene::TopDocsPtr hits = indexSearcherLocal->search(query, Lucene::FilterPtr(), maxResults);
				Lucene::Collection<Lucene::ScoreDocPtr>::iterator docIt;
				for (docIt = hits->scoreDocs.begin(); docIt != hits->scoreDocs.end(); docIt++ ) {
					xmlNodePtr docEl = addChildToResponse(xmlNewNode(NULL, BAD_CAST "docId"));
					xmlNewProp(docEl, BAD_CAST "score", BAD_CAST boost::lexical_cast<std::string>((*docIt)->score).c_str());

					Lucene::DocumentPtr foundDoc = indexSearcherLocal->doc((*docIt)->doc);
					Lucene::String id = foundDoc->get(L"id");
					xmlAddChild(docEl, xmlNewText(BAD_CAST convertToMultiByte(id.c_str())));
				}
			}
			catch (exception& pe) {
				syslog(LOG_ERR, "Error searching - %s", pe.what());
				addErrorToResponse("Error searching - " + string(pe.what()));
			}

			indexReaderLocal->decRef();
		}

		delete args;
	}
	else if (strncmp("POST", method, 4) == 0) {
		addErrorToResponse("Metodo POST non supportato!");
	}
	else {
		addErrorToResponse("Metodo " + string(method) + " non supportato!");
	}

	return this->toString();
}

}
