/*
 * Copyright 2012 Francesco Cappelli
 *
 * This file is part of Pagliaio.
 *
 * Pagliaio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pagliaio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Pagliaio.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <TreeIndexAction.h>
#include <Configuration.h>

extern "C" {
#include <syslog.h>
}

#include <lucene++/FileUtils.h>
#include <boost/lexical_cast.hpp>

using namespace std;

extern Lucene::IndexWriterPtr indexWriter;
extern pagliaio::Configuration configuration;

namespace pagliaio {

TreeIndexAction::TreeIndexAction() {}

TreeIndexAction::~TreeIndexAction() {}

string TreeIndexAction::generateResponse(void *cls, struct MHD_Connection *connection,
								 	 const char *url,
								 	 const char *method, const char *version,
								 	 const char *upload_data,
								 	 size_t *upload_data_size, void **con_cls)
{
	if (strncmp("GET", method, 3) == 0) {
		addErrorToResponse("Metodo GET non supportato!");
	}
	else if (strncmp("POST", method, 4) == 0) {
		openlog("pagliaio", LOG_ODELAY, LOG_DAEMON);

		if (*con_cls == NULL) {
			map<string, string>* args = new map<string, string>();
			MHD_PostProcessor* pProcessor = MHD_create_post_processor(connection, 512, iterate_post, (void*) args);

			PostSession* session = new PostSession();
			session->processor = pProcessor;
			session->data = args;
			*con_cls = (void*) session;

			return string("");
		}
		else {
			PostSession* session = (PostSession*)*con_cls;

			if (*upload_data_size != 0) {
				MHD_post_process(session->processor, upload_data, *upload_data_size);
				*upload_data_size = 0;
				return string("");
			}

			//controlla se gli argomenti ricevuti sono corretti
			if (session->data->count("id") == 0 || session->data->count("document") == 0) {
				addErrorToResponse("Numero o tipo degli argomenti errato.");
			}
			else {
				string idStr = string((*session->data)["id"]);
				string document = string((*session->data)["document"]);

				Lucene::DocumentPtr luceneDoc = Lucene::newLucene<Lucene::Document>();

				//indicizza tutti gli elementi
				//TODO aggiungere navigazione per xml multi livello
				xmlParserCtxtPtr ctxt = xmlNewParserCtxt();
				xmlSetGenericErrorFunc(ctxt, parseErrorFunc);

				xmlDocPtr incomingDoc;
				incomingDoc = xmlCtxtReadMemory(ctxt, document.c_str(), document.size(), std::string(idStr + ".xml").c_str(), NULL, XML_PARSE_NOBLANKS /*| XML_PARSE_NOERROR | XML_PARSE_NOWARNING*/);

				xmlErrorPtr error = xmlGetLastError();

				if (error != 0) {
					addErrorToResponse(std::string("Errore durante il parsing: ") + std::string(error->message) + " (riga " + boost::lexical_cast<std::string>(error->line) + ").");
					xmlFreeParserCtxt(ctxt);
					return this->toString();
				}

				xmlNodePtr root = xmlDocGetRootElement(incomingDoc);

				if (root == 0) {
					addErrorToResponse("Il documento è malformato: nessuna radice.");
					xmlFreeParserCtxt(ctxt);
					return this->toString();
				}

				string elementName = "/" + boost::lexical_cast<string>(root->name);
				string elementNameWNs = "/" + boost::lexical_cast<string>(root->name);

				if (root->ns != 0 && root->ns->prefix != 0) {
					elementNameWNs = "/" + boost::lexical_cast<std::string>(root->ns->prefix) + "_" + boost::lexical_cast<string>(root->name);
				}

				xmlNodePtr child = root->xmlChildrenNode;
				while (child != NULL) {
					parseAndIndexElement(luceneDoc, incomingDoc, child, elementName, elementNameWNs);

					child = child->next;
				}

				//libera la memoria del documento xml
				xmlFreeDoc(incomingDoc);
				xmlFreeParserCtxt(ctxt);

				//aggiungi l'id del documento
				Lucene::String id = Lucene::StringUtils::toUnicode(idStr.c_str());
				Lucene::FieldPtr field = Lucene::newLucene<Lucene::Field>(L"id", id, Lucene::Field::STORE_YES, Lucene::Field::INDEX_NOT_ANALYZED);
				luceneDoc->add(field);

				//salva il documento nell'indice
				indexWriter->addDocument(luceneDoc);
				indexWriter->commit();

				syslog(LOG_INFO, "Added document with id %s", idStr.c_str());

				xmlNodePtr indexedDocs = addChildToResponse(xmlNewNode(NULL, BAD_CAST "indexedDocs"));

				stringstream ss;
				int docCount = 0;
				for (int i=0; i < indexWriter->getSegmentCount(); i++) {
					docCount += indexWriter->getDocCount(i);
				}
				ss << docCount;
				xmlAddChild(indexedDocs, xmlNewText(BAD_CAST ss.str().c_str()));
			}

			MHD_destroy_post_processor(session->processor);
			delete session->data;
			delete session;

			return toString();
		}
	}
	else {
		addErrorToResponse("Metodo " + string(method) + " non supportato!");
	}

	return this->toString();
}

void TreeIndexAction::parseAndIndexElement(Lucene::DocumentPtr luceneDoc, xmlDocPtr incomingDoc, xmlNodePtr node, string parentPath, string parentPathWNs) {
	if (node != 0) {
		switch(node->type) {
		case XML_ELEMENT_NODE:
		{
			string elementName = parentPath + "/" + boost::lexical_cast<std::string>(node->name);

			Lucene::String key = Lucene::StringUtils::toUnicode(elementName.c_str());

			char* elementText = (char*)xmlNodeListGetString(incomingDoc, node->xmlChildrenNode, 1);

			Lucene::String value;
			if (elementText != 0) {
				value = Lucene::StringUtils::toUnicode(elementText);
				xmlFree(elementText);
			}

			Lucene::FieldPtr field = Lucene::newLucene<Lucene::Field>(key, value, configuration.indexNeedStoring(elementName), configuration.getIndexType(elementName));
			luceneDoc->add(field);

			//se contiene un namespace aggiungilo e indicizza anche con chiave namespace_elementName
			xmlNsPtr ns = node->ns;

			string elementNameWNs = parentPathWNs + "/" + boost::lexical_cast<std::string>(node->name);
			if (ns != 0 && ns->prefix != 0) {
				//rimpiazza il carattere : con un underscore per gestire namespace e non far impazzire lucene
				elementNameWNs = parentPathWNs + "/" + boost::lexical_cast<std::string>(ns->prefix) + "_" + boost::lexical_cast<std::string>(node->name);
				key = Lucene::StringUtils::toUnicode(elementNameWNs.c_str());
				field = Lucene::newLucene<Lucene::Field>(key, value, configuration.indexNeedStoring(elementNameWNs), configuration.getIndexType(elementNameWNs));
				luceneDoc->add(field);
			}

			//indicizza gli attributi...
			xmlNodePtr attribute = (xmlNodePtr)((xmlElementPtr)node)->attributes;
			while(attribute != 0) {
				parseAndIndexElement(luceneDoc, incomingDoc, attribute, elementName, elementNameWNs);
				attribute = attribute->next;
			}

			//... e poi fatti i figli
			xmlNodePtr child = node->xmlChildrenNode;
			while (child != NULL) {
				parseAndIndexElement(luceneDoc, incomingDoc, child, elementName, elementNameWNs);
				child = child->next;
			}
			break;
		}
		case XML_ATTRIBUTE_NODE:
		{
			string elementName = parentPath + "/@" + boost::lexical_cast<std::string>(node->name);
			char* elementText = (char*)xmlNodeListGetString(incomingDoc, node->xmlChildrenNode, 1);

			Lucene::String key = Lucene::StringUtils::toUnicode(elementName.c_str());

			Lucene::String value;
			if (elementText != 0)
				value = Lucene::StringUtils::toUnicode(elementText);
			else
				value = Lucene::StringUtils::toUnicode("");

			xmlFree(elementText);

			Lucene::FieldPtr field = Lucene::newLucene<Lucene::Field>(key, value, configuration.indexNeedStoring(elementName), configuration.getIndexType(elementName));
			luceneDoc->add(field);

			break;
		}
		case XML_TEXT_NODE:
		{
			break;
		}
		case XML_CDATA_SECTION_NODE:
		{
			break;
		}
		case XML_ENTITY_REF_NODE:
		case XML_ENTITY_NODE:
		case XML_PI_NODE:
		case XML_COMMENT_NODE:
		case XML_DOCUMENT_NODE:
		case XML_DOCUMENT_TYPE_NODE:
		case XML_DOCUMENT_FRAG_NODE:
		case XML_NOTATION_NODE:
		case XML_HTML_DOCUMENT_NODE:
		case XML_DTD_NODE:
		case XML_ELEMENT_DECL:
		case XML_ATTRIBUTE_DECL:
		case XML_ENTITY_DECL:
		case XML_NAMESPACE_DECL:
		case XML_XINCLUDE_START:
		case XML_XINCLUDE_END:
		case XML_DOCB_DOCUMENT_NODE:
		default:

			break;
		}
	}
}

}
